@@ -4,18 +4,21 @@ mod tests;
 
 use crate::directory::Directory;
 use crate::run::PathDependency;
+use std::path::Path;
 
 #[derive(Copy, Clone)]
 pub struct Context<'a> {
     pub krate: &'a str,
     pub source_dir: &'a Directory,
     pub workspace: &'a Directory,
+    pub input_file: &'a Path,
+    pub target_dir: &'a Directory,
     pub path_dependencies: &'a [PathDependency],
 }
 
 pub fn trim<S: AsRef<[u8]>>(output: S) -> String {
     let bytes = output.as_ref();
-    let mut normalized = String::from_utf8_lossy(bytes).to_string();
+    let mut normalized = String::from_utf8_lossy(bytes).into_owned();
 
     let len = normalized.trim_end().len();
     normalized.truncate(len);
@@ -38,9 +41,8 @@ pub fn trim<S: AsRef<[u8]>>(output: S) -> String {
 ///
 /// There is one "preferred" variation which is what we print when the stderr
 /// file is absent or not a match.
-pub fn diagnostics(output: Vec<u8>, context: Context) -> Variations {
-    let mut from_bytes = String::from_utf8_lossy(&output).to_string();
-    from_bytes = from_bytes.replace("\r\n", "\n");
+pub fn diagnostics(output: &str, context: Context) -> Variations {
+    let output = output.replace("\r\n", "\n");
 
     let variations = [
         Basic,
@@ -56,9 +58,10 @@ pub fn diagnostics(output: Vec<u8>, context: Context) -> Variations {
         CargoRegistry,
         ArrowOtherCrate,
         RelativeToDir,
+        LinesOutsideInputFile,
     ]
     .iter()
-    .map(|normalization| apply(&from_bytes, *normalization, context))
+    .map(|normalization| apply(&output, *normalization, context))
     .collect();
 
     Variations { variations }
@@ -93,6 +96,7 @@ enum Normalization {
     CargoRegistry,
     ArrowOtherCrate,
     RelativeToDir,
+    LinesOutsideInputFile,
     // New normalization steps are to be inserted here at the end so that any
     // snapshots saved before your normalization change remain passing.
 }
@@ -158,30 +162,80 @@ impl<'a> Filter<'a> {
         if let Some(prefix) = prefix {
             line = line.replace('\\', "/");
             let line_lower = line.to_ascii_lowercase();
+            let target_dir_pat = self
+                .context
+                .target_dir
+                .to_string_lossy()
+                .to_ascii_lowercase()
+                .replace('\\', "/");
             let source_dir_pat = self
                 .context
                 .source_dir
                 .to_string_lossy()
                 .to_ascii_lowercase()
                 .replace('\\', "/");
-            if let Some(i) = line_lower.find(&source_dir_pat) {
+            let mut other_crate = false;
+            if line_lower.find(&target_dir_pat) == Some(indent + 4) {
+                let mut offset = indent + 4 + target_dir_pat.len();
+                let mut out_dir_crate_name = None;
+                while let Some(slash) = line[offset..].find('/') {
+                    let component = &line[offset..offset + slash];
+                    if component == "out" {
+                        if let Some(out_dir_crate_name) = out_dir_crate_name {
+                            let replacement = format!("$OUT_DIR[{}]", out_dir_crate_name);
+                            line.replace_range(indent + 4..offset + 3, &replacement);
+                            other_crate = true;
+                            break;
+                        }
+                    } else if component.len() > 17
+                        && component.rfind('-') == Some(component.len() - 17)
+                        && component[component.len() - 16..].bytes().all(|b| match b {
+                            b'0'..=b'9' | b'a'..=b'f' => true,
+                            _ => false,
+                        })
+                    {
+                        out_dir_crate_name = Some(&component[..component.len() - 17]);
+                    } else {
+                        out_dir_crate_name = None;
+                    }
+                    offset += slash + 1;
+                }
+            } else if let Some(i) = line_lower.find(&source_dir_pat) {
                 if self.normalization >= RelativeToDir && i == indent + 4 {
                     line.replace_range(i..i + source_dir_pat.len(), "");
+                    if self.normalization < LinesOutsideInputFile {
+                        return Some(line);
+                    }
+                    let input_file_pat = self
+                        .context
+                        .input_file
+                        .to_string_lossy()
+                        .to_ascii_lowercase()
+                        .replace('\\', "/");
+                    if line_lower[i + source_dir_pat.len()..].starts_with(&input_file_pat) {
+                        // Keep line numbers only within the input file (the
+                        // path passed to our `fn compile_fail`. All other
+                        // source files get line numbers erased below.
+                        return Some(line);
+                    }
                 } else {
                     line.replace_range(i..i + source_dir_pat.len() - 1, "$DIR");
+                    if self.normalization < LinesOutsideInputFile {
+                        return Some(line);
+                    }
                 }
-                return Some(line);
-            }
-            let mut other_crate = false;
-            let workspace_pat = self
-                .context
-                .workspace
-                .to_string_lossy()
-                .to_ascii_lowercase()
-                .replace('\\', "/");
-            if let Some(i) = line_lower.find(&workspace_pat) {
-                line.replace_range(i..i + workspace_pat.len() - 1, "$WORKSPACE");
                 other_crate = true;
+            } else {
+                let workspace_pat = self
+                    .context
+                    .workspace
+                    .to_string_lossy()
+                    .to_ascii_lowercase()
+                    .replace('\\', "/");
+                if let Some(i) = line_lower.find(&workspace_pat) {
+                    line.replace_range(i..i + workspace_pat.len() - 1, "$WORKSPACE");
+                    other_crate = true;
+                }
             }
             if self.normalization >= PathDependencies && !other_crate {
                 for path_dep in self.context.path_dependencies {
@@ -326,7 +380,13 @@ fn replace_case_insensitive(line: &str, pattern: &str, replacement: &str) -> Str
             replaced.push_str(replacement);
             pos += pattern.len();
         }
-        let keep = &line[pos..pos + keep.len()];
+        let mut keep = &line[pos..pos + keep.len()];
+        if insert_replacement {
+            let end_of_maybe_path = keep.find(&[' ', ':'][..]).unwrap_or(keep.len());
+            replaced.push_str(&keep[..end_of_maybe_path].replace('\\', "/"));
+            pos += end_of_maybe_path;
+            keep = &keep[end_of_maybe_path..];
+        }
         replaced.push_str(keep);
         pos += keep.len();
         insert_replacement = true;
